@@ -5,22 +5,20 @@
 # NIPALS
 pls_nipals <- function(x, y, k = 3L, center = TRUE, scale. = FALSE,
 	transpose = FALSE, niter = 100L, tol = 1e-5,
-	verbose = NA, nchunks = NA, BPPARAM = bpparam(), ...)
+	verbose = NA, BPPARAM = bpparam(), ...)
 {
 	if ( is.na(verbose) )
 		verbose <- getOption("matter.default.verbose")
-	if ( is.na(nchunks) )
-		nchunks <- getOption("matter.default.nchunks")
 	x <- as_real_memory_matrix(x)
 	k <- min(max(k), dim(x))
 	# center and scale x and y
-	if ( verbose && (center || scale.) )
-		message("preparing x and y matrices")
+	if ( center || scale. )
+		matter_log("preparing x and y matrices", verbose=verbose)
 	if ( is.factor(y) || is.character(y) )
 		y <- encode_dummy(y)
 	y <- as.matrix(y)
 	y <- colscale(y, center=center, scale=scale.,
-		verbose=verbose, nchunks=nchunks, BPPARAM=BPPARAM)
+		verbose=verbose, BPPARAM=BPPARAM, ...)
 	y.center <- attr(y, "col-scaled:center")
 	y.scale <- attr(y, "col-scaled:scale")
 	if ( transpose ) {
@@ -29,7 +27,7 @@ pls_nipals <- function(x, y, k = 3L, center = TRUE, scale. = FALSE,
 		pnames <- rownames(x)
 		snames <- colnames(x)
 		xt <- rowscale(x, center=center, scale=scale.,
-			verbose=verbose, nchunks=nchunks, BPPARAM=BPPARAM)
+			verbose=verbose, BPPARAM=BPPARAM, ...)
 		center <- attr(xt, "row-scaled:center")
 		scale <- attr(xt, "row-scaled:scale")
 		xt0 <- xt
@@ -39,7 +37,7 @@ pls_nipals <- function(x, y, k = 3L, center = TRUE, scale. = FALSE,
 		pnames <- colnames(x)
 		snames <- rownames(x)
 		x <- colscale(x, center=center, scale=scale.,
-			verbose=verbose, nchunks=nchunks, BPPARAM=BPPARAM)
+			verbose=verbose, BPPARAM=BPPARAM, ...)
 		center <- attr(x, "col-scaled:center")
 		scale <- attr(x, "col-scaled:scale")
 		x0 <- x
@@ -56,13 +54,12 @@ pls_nipals <- function(x, y, k = 3L, center = TRUE, scale. = FALSE,
 		dimnames=list(colnames(y), paste0("C", j)))
 	y.scores <- matrix(nrow=nrow(y), ncol=k,
 		dimnames=list(rownames(y), paste0("C", j)))
-	cvar <- setNames(numeric(k), paste0("C", j))
+	cvar <- set_names(numeric(k), paste0("C", j))
 	inner <- numeric(k)
 	y0 <- y
 	for ( i in j )
 	{
-		if ( verbose )
-			message("fitting PLS component ", i)
+		matter_log("fitting PLS component ", i, verbose=verbose)
 		u <- y[,which.max(colSums(y)),drop=FALSE]
 		du <- Inf
 		iter <- 1
@@ -87,7 +84,7 @@ pls_nipals <- function(x, y, k = 3L, center = TRUE, scale. = FALSE,
 			iter <- iter + 1
 		}
 		if ( iter > niter && du > tol )
-			warning("NIPALS did not converge in ",
+			matter_warn("NIPALS did not converge in ",
 				iter - 1, " iterations for component ", i)
 		# deflate x and y based on projection
 		if ( transpose ) {
@@ -111,8 +108,7 @@ pls_nipals <- function(x, y, k = 3L, center = TRUE, scale. = FALSE,
 		inner[i] <- c
 	}
 	# calculate regression coefficients
-	if ( verbose )
-		message("fitting regression coefficients")
+	matter_log("fitting regression coefficients", verbose=verbose)
 	projection <- weights %*% solve(crossprod(loadings, weights))
 	projection <- projection %*% (inner * diag(k))
 	dimnames(projection) <- list(pnames, paste0("C", j))
@@ -143,21 +139,19 @@ pls_nipals <- function(x, y, k = 3L, center = TRUE, scale. = FALSE,
 # SIMPLS
 pls_simpls <- function(x, y, k = 3L, center = TRUE, scale. = FALSE,
 	transpose = FALSE, method = 1L, retscores = TRUE,
-	verbose = NA, nchunks = NA, BPPARAM = bpparam(), ...)
+	verbose = NA, BPPARAM = bpparam(), ...)
 {
 	if ( is.na(verbose) )
 		verbose <- getOption("matter.default.verbose")
-	if ( is.na(nchunks) )
-		nchunks <- getOption("matter.default.nchunks")
 	k <- min(max(k), dim(x))
 	# center and scale x and y + calculate covariance
-	if ( verbose && (center || scale.) )
-		message("preparing x and y matrices")
+	if ( center || scale. )
+		matter_log("preparing x and y matrices", verbose=verbose)
 	if ( is.factor(y) || is.character(y) )
 		y <- encode_dummy(y)
 	y <- as.matrix(y)
 	y <- colscale(y, center=center, scale=scale.,
-		verbose=verbose, nchunks=nchunks, BPPARAM=BPPARAM)
+		verbose=verbose, BPPARAM=BPPARAM, ...)
 	y.center <- attr(y, "col-scaled:center")
 	y.scale <- attr(y, "col-scaled:scale")
 	if ( transpose ) {
@@ -166,7 +160,7 @@ pls_simpls <- function(x, y, k = 3L, center = TRUE, scale. = FALSE,
 		pnames <- rownames(x)
 		snames <- colnames(x)
 		xt <- rowscale(x, center=center, scale=scale.,
-			verbose=verbose, nchunks=nchunks, BPPARAM=BPPARAM)
+			verbose=verbose, BPPARAM=BPPARAM, ...)
 		center <- attr(xt, "row-scaled:center")
 		scale <- attr(xt, "row-scaled:scale")
 		xy <- xt %*% y
@@ -176,7 +170,7 @@ pls_simpls <- function(x, y, k = 3L, center = TRUE, scale. = FALSE,
 		pnames <- colnames(x)
 		snames <- rownames(x)
 		x <- colscale(x, center=center, scale=scale.,
-			verbose=verbose, nchunks=nchunks, BPPARAM=BPPARAM)
+			verbose=verbose, BPPARAM=BPPARAM, ...)
 		center <- attr(x, "col-scaled:center")
 		scale <- attr(x, "col-scaled:scale")
 		xy <- crossprod(x, y)
@@ -195,13 +189,12 @@ pls_simpls <- function(x, y, k = 3L, center = TRUE, scale. = FALSE,
 			dimnames=list(snames, paste0("C", j)))
 		y.scores <- matrix(nrow=nrow(y), ncol=k,
 			dimnames=list(rownames(y), paste0("C", j)))
-		cvar <- setNames(numeric(k), paste0("C", j))
+		cvar <- set_names(numeric(k), paste0("C", j))
 	}
 	for ( i in j )
 	{
 		# show progress
-		if ( verbose )
-			message("fitting PLS component ", i)
+		matter_log("fitting PLS component ", i, verbose=verbose)
 		# calculate scores and loadings
 		s <- svd(xy)
 		r <- s$u[,1L,drop=FALSE]
@@ -240,8 +233,7 @@ pls_simpls <- function(x, y, k = 3L, center = TRUE, scale. = FALSE,
 		}
 	}
 	# calculate regression coefficients
-	if ( verbose )
-		message("fitting regression coefficients")
+	matter_log("fitting regression coefficients", verbose=verbose)
 	b <- tcrossprod(projection, y.loadings)
 	if ( transpose ) {
 		yhat <- t(t(b) %*% xt)
@@ -275,21 +267,19 @@ pls_simpls <- function(x, y, k = 3L, center = TRUE, scale. = FALSE,
 # Kernel (#1 and #2)
 pls_kernel <- function(x, y, k = 3L, center = TRUE, scale. = FALSE,
 	transpose = FALSE, method = 1L, retscores = TRUE,
-	verbose = NA, nchunks = NA, BPPARAM = bpparam(), ...)
+	verbose = NA, BPPARAM = bpparam(), ...)
 {
 	if ( is.na(verbose) )
 		verbose <- getOption("matter.default.verbose")
-	if ( is.na(nchunks) )
-		nchunks <- getOption("matter.default.nchunks")
 	k <- min(max(k), dim(x))
 	# center and scale x and y + calculate covariance
-	if ( verbose && (center || scale.) )
-		message("preparing x and y matrices")
+	if ( center || scale. )
+		matter_log("preparing x and y matrices", verbose=verbose)
 	if ( is.factor(y) || is.character(y) )
 		y <- encode_dummy(y)
 	y <- as.matrix(y)
 	y <- colscale(y, center=center, scale=scale.,
-		verbose=verbose, nchunks=nchunks, BPPARAM=BPPARAM)
+		verbose=verbose, BPPARAM=BPPARAM, ...)
 	y.center <- attr(y, "col-scaled:center")
 	y.scale <- attr(y, "col-scaled:scale")
 	if ( transpose ) {
@@ -298,7 +288,7 @@ pls_kernel <- function(x, y, k = 3L, center = TRUE, scale. = FALSE,
 		pnames <- rownames(x)
 		snames <- colnames(x)
 		xt <- rowscale(x, center=center, scale=scale.,
-			verbose=verbose, nchunks=nchunks, BPPARAM=BPPARAM)
+			verbose=verbose, BPPARAM=BPPARAM, ...)
 		center <- attr(xt, "row-scaled:center")
 		scale <- attr(xt, "row-scaled:scale")
 		xy <- xt %*% y
@@ -310,7 +300,7 @@ pls_kernel <- function(x, y, k = 3L, center = TRUE, scale. = FALSE,
 		pnames <- colnames(x)
 		snames <- rownames(x)
 		x <- colscale(x, center=center, scale=scale.,
-			verbose=verbose, nchunks=nchunks, BPPARAM=BPPARAM)
+			verbose=verbose, BPPARAM=BPPARAM, ...)
 		center <- attr(x, "col-scaled:center")
 		scale <- attr(x, "col-scaled:scale")
 		xy <- crossprod(x, y)
@@ -332,13 +322,12 @@ pls_kernel <- function(x, y, k = 3L, center = TRUE, scale. = FALSE,
 			dimnames=list(snames, paste0("C", j)))
 		y.scores <- matrix(nrow=nrow(y), ncol=k,
 			dimnames=list(rownames(y), paste0("C", j)))
-		cvar <- setNames(numeric(k), paste0("C", j))
+		cvar <- set_names(numeric(k), paste0("C", j))
 	}
 	for ( i in j )
 	{
 		# show progress
-		if ( verbose )
-			message("fitting PLS component ", i)
+		matter_log("fitting PLS component ", i, verbose=verbose)
 		# calculate projection vectors
 		if ( ncol(y) < P ) {
 			yxxy <- crossprod(xy, xy)
@@ -379,7 +368,7 @@ pls_kernel <- function(x, y, k = 3L, center = TRUE, scale. = FALSE,
 			p <- crossprod(xx, r) / tt
 			q <- crossprod(xy, r) / tt
 		} else {
-			stop("unrecognized kernel method: ", method)
+			matter_error("unrecognized kernel method: ", method)
 		}
 		# update covariance
 		xy <- xy - tcrossprod(p, q) * tt
@@ -400,8 +389,7 @@ pls_kernel <- function(x, y, k = 3L, center = TRUE, scale. = FALSE,
 		}
 	}
 	# calculate regression coefficients
-	if ( verbose )
-		message("fitting regression coefficients")
+	matter_log("fitting regression coefficients", verbose=verbose)
 	b <- tcrossprod(projection, y.loadings)
 	if ( transpose ) {
 		yhat <- t(t(b) %*% xt)
@@ -449,17 +437,17 @@ predict.pls <- function(object, newdata, k,
 	if ( missing(newdata) && missing(k) )
 		return(fitted(object, type=type))
 	if ( missing(newdata) )
-		stop("'newdata' must be specified if 'k' is specified")
+		matter_error("'newdata' must be specified if 'k' is specified")
 	if ( missing(k) )
 		k <- ncol(object$loadings)
 	if ( length(dim(newdata)) != 2L )
-		stop("'newdata' must be a matrix or data frame")
+		matter_error("'newdata' must be a matrix or data frame")
 	nm <- rownames(object$loadings)
 	v <- if (object$transpose) rownames(newdata) else colnames(newdata)
 	p <- if (object$transpose) nrow(newdata) else ncol(newdata)
 	if ( !is.null(nm) ) {
 		if ( !all(nm %in% v) )
-			stop("'newdata' does not have named features ",
+			matter_error("'newdata' does not have named features ",
 				"matching one of more of the original features")
 		if ( object$transpose ) {
 			newdata <- newdata[nm,,drop=FALSE]
@@ -468,7 +456,7 @@ predict.pls <- function(object, newdata, k,
 		}
 	} else {
 		if ( p != nrow(object$loadings) )
-			stop("'newdata' does not have the correct number of features")
+			matter_error("'newdata' does not have the correct number of features")
 	}
 	if ( object$transpose ) {
 		xt <- rowscale(newdata, center=object$center, scale=object$scale)
@@ -477,9 +465,9 @@ predict.pls <- function(object, newdata, k,
 	}
 	
 	if ( any(k > ncol(object$loadings)) )
-		stop("'k' is larger than the number of components")
+		matter_error("'k' is larger than the number of components")
 	# predict for each k
-	ans <- lapply(setNames(k, paste0("k=", k)),
+	ans <- lapply(set_names(k, paste0("k=", k)),
 		function(ki)
 		{
 			# extract relevant components
@@ -521,7 +509,7 @@ print.pls <- function(x, ...)
 	cat(sprintf("Partial least squares (k=%d)\n", ncol(x$loadings)))
 	if ( !is.null(x$cvar) ) {
 		cat(sprintf("\nCovariances (1, .., k=%d):\n", length(x$cvar)))
-	    preview_vector(x$cvar, ...)
+		preview_vector(x$cvar, ...)
 	}
 	if ( !is.null(coef(x)) ) {
 		cat("\nCoefficients:\n")
@@ -537,14 +525,14 @@ vip <- function(object, type = c("projection", "weights"))
 	type <- match.arg(type)
 	w0 <- object[[type]]
 	if ( is.null(w0) )
-		stop("missing component: ", sQuote(type))
+		matter_error("missing component: ", sQuote(type))
 	P <- nrow(w0)
 	scores <- object$scores
 	y.loadings <- object$y.loadings
 	if ( is.null(scores) )
-		stop("missing component: 'scores'")
+		matter_error("missing component: 'scores'")
 	if ( is.null(y.loadings) )
-		stop("missing component: 'y.loadings'")
+		matter_error("missing component: 'y.loadings'")
 	w0 <- colsweep_matrix(w0, sqrt(colSums(w0^2)), "/")
 	ssy <- colSums(scores^2) * colSums(y.loadings^2)
 	vip <- sqrt(P * rowSums(ssy * w0^2) / sum(ssy))
@@ -558,23 +546,21 @@ vip <- function(object, type = c("projection", "weights"))
 # NIPALS
 opls_nipals <- function(x, y, k = 3L, center = TRUE, scale. = FALSE,
 	transpose = FALSE, niter = 100L, tol = 1e-9, regression = TRUE,
-	verbose = NA, nchunks = NA, BPPARAM = bpparam(), ...)
+	verbose = NA, BPPARAM = bpparam(), ...)
 {
 	if ( is.na(verbose) )
 		verbose <- getOption("matter.default.verbose")
-	if ( is.na(nchunks) )
-		nchunks <- getOption("matter.default.nchunks")
 	y_in <- y
 	x <- as_real_memory_matrix(x)
 	k <- min(max(k), dim(x))
 	# center and scale x and y
-	if ( verbose && (center || scale.) )
-		message("preparing x and y matrices")
+	if ( center || scale. )
+		matter_log("preparing x and y matrices", verbose=verbose)
 	if ( is.factor(y) || is.character(y) )
 		y <- encode_dummy(y)
 	y <- as.matrix(y)
 	y <- colscale(y, center=center, scale=scale.,
-		verbose=verbose, nchunks=nchunks, BPPARAM=BPPARAM)
+		verbose=verbose, BPPARAM=BPPARAM, ...)
 	y.center <- attr(y, "col-scaled:center")
 	y.scale <- attr(y, "col-scaled:scale")
 	if ( transpose ) {
@@ -583,7 +569,7 @@ opls_nipals <- function(x, y, k = 3L, center = TRUE, scale. = FALSE,
 		pnames <- rownames(x)
 		snames <- colnames(x)
 		xt <- rowscale(x, center=center, scale=scale.,
-			verbose=verbose, nchunks=nchunks, BPPARAM=BPPARAM)
+			verbose=verbose, BPPARAM=BPPARAM, ...)
 		center <- attr(xt, "row-scaled:center")
 		scale <- attr(xt, "row-scaled:scale")
 		xt0 <- xt
@@ -593,7 +579,7 @@ opls_nipals <- function(x, y, k = 3L, center = TRUE, scale. = FALSE,
 		pnames <- colnames(x)
 		snames <- rownames(x)
 		x <- colscale(x, center=center, scale=scale.,
-			verbose=verbose, nchunks=nchunks, BPPARAM=BPPARAM)
+			verbose=verbose, BPPARAM=BPPARAM, ...)
 		center <- attr(x, "col-scaled:center")
 		scale <- attr(x, "col-scaled:scale")
 		x0 <- x
@@ -608,7 +594,7 @@ opls_nipals <- function(x, y, k = 3L, center = TRUE, scale. = FALSE,
 		dimnames=list(pnames, paste0("C", j)))
 	scores <- matrix(nrow=N, ncol=k,
 		dimnames=list(snames, paste0("C", j)))
-	ratio <- setNames(numeric(k), paste0("C", j))
+	ratio <- set_names(numeric(k), paste0("C", j))
 	y0 <- y
 	if ( transpose ) {
 		w0 <- colsweep_matrix((xt %*% y), colSums(y^2), "/")
@@ -619,8 +605,7 @@ opls_nipals <- function(x, y, k = 3L, center = TRUE, scale. = FALSE,
 	for ( i in j )
 	{
 		# show progress
-		if ( verbose )
-			message("fitting OPLS component ", i)
+		matter_log("fitting OPLS component ", i, verbose=verbose)
 		u <- y[,which.max(colSums(y)),drop=FALSE]
 		du <- Inf
 		iter <- 1
@@ -645,7 +630,7 @@ opls_nipals <- function(x, y, k = 3L, center = TRUE, scale. = FALSE,
 			iter <- iter + 1
 		}
 		if ( iter > niter && du > tol )
-			warning("NIPALS did not converge in ",
+			matter_warn("NIPALS did not converge in ",
 				iter - 1, " iterations for component ", i)
 		# orthogonalize projection
 		if ( transpose ) {
@@ -663,8 +648,7 @@ opls_nipals <- function(x, y, k = 3L, center = TRUE, scale. = FALSE,
 				regressions[[i]] <- pls_nipals(xt, y=y_in, k=1L,
 					center=FALSE, scale.=FALSE,
 					transpose=TRUE, niter=niter, tol=tol,
-					nchunks=nchunks, verbose=verbose,
-					BPPARAM=BPPARAM, ...)
+					verbose=verbose, BPPARAM=BPPARAM, ...)
 			}
 		} else {
 			p <- crossprod(x, t) / tt
@@ -681,8 +665,7 @@ opls_nipals <- function(x, y, k = 3L, center = TRUE, scale. = FALSE,
 				regressions[[i]] <- pls_nipals(x, y=y_in, k=1L,
 					center=FALSE, scale.=FALSE,
 					transpose=FALSE, niter=niter, tol=tol,
-					nchunks=nchunks, verbose=verbose,
-					BPPARAM=BPPARAM, ...)
+					verbose=verbose, BPPARAM=BPPARAM, ...)
 			}
 		}
 		weights[,i] <- wo
@@ -708,7 +691,7 @@ opls_nipals <- function(x, y, k = 3L, center = TRUE, scale. = FALSE,
 coef.opls <- function(object, ...)
 {
 	if ( is.null(object$regressions) )
-		stop("missing component: 'regressions'")
+		matter_error("missing component: 'regressions'")
 	k <- length(object$regressions)
 	coef(object$regressions[[k]], ...)
 }
@@ -716,7 +699,7 @@ coef.opls <- function(object, ...)
 residuals.opls <- function(object, ...)
 {
 	if ( is.null(object$regressions) )
-		stop("missing component: 'regressions'")
+		matter_error("missing component: 'regressions'")
 	k <- length(object$regressions)
 	residuals(object$regressions[[k]], ...)
 }
@@ -726,7 +709,7 @@ fitted.opls <- function(object, type = c("response", "class", "x"), ...)
 	type <- match.arg(type)
 	if ( type != "x" ) {
 		if ( is.null(object$regressions) )
-			stop("missing component: 'regressions'")
+			matter_error("missing component: 'regressions'")
 		k <- length(object$regressions)
 		fitted(object$regressions[[k]], type=type, ...)
 	} else {
@@ -741,19 +724,19 @@ predict.opls <- function(object, newdata, k,
 	if ( missing(newdata) && missing(k) )
 		return(fitted(object, type=type))
 	if ( type != "x" && is.null(object$regressions) )
-		stop("missing component: 'regressions'")
+		matter_error("missing component: 'regressions'")
 	if ( missing(newdata) )
-		stop("'newdata' must be specified if 'k' is specified")
+		matter_error("'newdata' must be specified if 'k' is specified")
 	if ( missing(k) )
 		k <- ncol(object$loadings)
 	if ( length(dim(newdata)) != 2L )
-		stop("'newdata' must be a matrix or data frame")
+		matter_error("'newdata' must be a matrix or data frame")
 	nm <- rownames(object$loadings)
 	v <- if (object$transpose) rownames(newdata) else colnames(newdata)
 	p <- if (object$transpose) nrow(newdata) else ncol(newdata)
 	if ( !is.null(nm) ) {
 		if ( !all(nm %in% v) )
-			stop("'newdata' does not have named features ",
+			matter_error("'newdata' does not have named features ",
 				"matching one of more of the original features")
 		if ( object$transpose ) {
 			newdata <- newdata[nm,,drop=FALSE]
@@ -762,12 +745,12 @@ predict.opls <- function(object, newdata, k,
 		}
 	} else {
 		if ( p != nrow(object$loadings) )
-			stop("'newdata' does not have the correct number of features")
+			matter_error("'newdata' does not have the correct number of features")
 	}
 	# extract relevant components
 	if ( any(k > ncol(object$loadings)) )
-		stop("'k' is larger than the number of components")
-	ans <- lapply(setNames(k, paste0("k=", k)),
+		matter_error("'k' is larger than the number of components")
+	ans <- lapply(set_names(k, paste0("k=", k)),
 		function(ki)
 		{
 			weights <- object$weights[,1:ki,drop=FALSE]
@@ -814,8 +797,8 @@ print.opls <- function(x, print.x = FALSE, ...)
 {
 	cat(sprintf("Orthogonal partial least squares (k=%d)\n", ncol(x$loadings)))
 	cat(sprintf("\nComponent ratios (1, .., k=%d):\n", length(x$ratio)))
-    preview_vector(x$ratio, ...)
-    if ( print.x && !is.null(x$x) ) {
+	preview_vector(x$ratio, ...)
+	if ( print.x && !is.null(x$x) ) {
 		cat("\nProcessed variables:\n")
 		preview_matrix(x$x, ...)
 	}
