@@ -17,7 +17,7 @@ setClass("sparse_arr",
 		sampler = "factor",
 		ops = "ops_OR_NULL",
 		transpose = "logical"),
-	contains = c("matter", "VIRTUAL"),
+	contains = "matter",
 	validity = function(object) {
 		errors <- NULL
 		if ( is.null(object@dim) )
@@ -220,9 +220,62 @@ as.sparse <- function(x, ...) sparse_mat(x, ...)
 
 is.sparse <- function(x) is(x, "sparse_arr")
 
-setMethod("as.matrix", "sparse_arr", function(x) x[])
+setAs("sparse_arr", "sparse_vec",
+	function(from) new("sparse_vec", from, dim=length(from), dimnames=NULL))
 
-setMethod("as.array", "sparse_arr", function(x) x[])
+setAs("sparse_mat", "sparse_vec",
+	function(from) new("sparse_vec", from, dim=length(from), dimnames=NULL))
+
+setAs("sparse_arr", "sparse_mat",
+	function(from) {
+		x <- new("sparse_mat", from)
+		if ( validObject(x) )
+			x
+	})
+
+setAs("sparse_vec", "sparse_mat",
+	function(from) {
+		dm <- c(length(from), 1L)
+		x <- new("sparse_mat", from, dim=dm, names=NULL)
+		if ( validObject(x) )
+			x
+	})
+
+setMethod("as.vector", "sparse_arr",
+	function(x, mode = "any") {
+		if ( mode != "any" )
+			type(x) <- mode
+		names(x) <- NULL
+		dimnames(x) <- NULL
+		as(x, "sparse_vec")[]
+	})
+
+setMethod("as.raw", "sparse_arr",
+	function(x) as.vector(x, "raw"))
+
+setMethod("as.logical", "sparse_arr",
+	function(x, ...) as.vector(x, "logical"))
+
+setMethod("as.integer", "sparse_arr",
+	function(x, ...) as.vector(x, "integer"))
+
+setMethod("as.double", "sparse_arr",
+	function(x, ...) as.vector(x, "double"))
+
+setMethod("as.numeric", "sparse_arr",
+	function(x, ...) as.vector(x, "double"))
+
+setAs("sparse_arr", "matrix", function(from) as.matrix(from[]))
+
+setAs("sparse_arr", "array", function(from) as.array(from[]))
+
+as.matrix.sparse_arr <- function(x, ...) as(x, "matrix")
+
+as.array.sparse_arr <- function(x, ...) as(x, "array")
+
+setMethod("as.matrix", "sparse_arr", as.matrix.sparse_arr)
+
+setMethod("as.array", "sparse_arr", as.array.sparse_arr)
 
 setMethod("describe_for_display", "sparse_mat", function(x) {
 	desc1 <- paste0("<", nrow(x), " row x ", ncol(x), " col> ", class(x))
@@ -283,6 +336,32 @@ setMethod("vm_used", "sparse_arr", function(x) {
 	size_bytes(sum(c(vm_index, vm_data), na.rm=TRUE))
 })
 
+setMethod("shm_used", "sparse_arr", function(x) {
+	shm_index <- shm_used(atomindex(x))
+	shm_data <- shm_used(atomdata(x))
+	size_bytes(sum(c(shm_index, shm_data), na.rm=TRUE))
+})
+
+setMethod("mem_realized", "sparse_arr", function(x) {
+	vm_index <- mem_realized(atomindex(x))
+	vm_data <- mem_realized(atomdata(x))
+	size_bytes(sum(c(vm_index, vm_data), na.rm=TRUE))
+})
+
+setMethod("fetch", "sparse_arr",
+	function(object, ..., BPPARAM = bpparam()) {
+		atomindex(object) <- fetch(atomindex(object), ..., BPPARAM=BPPARAM)
+		atomdata(object) <- fetch(atomdata(object), ..., BPPARAM=BPPARAM)
+		object
+	})
+
+setMethod("flash", "sparse_arr",
+	function(object, ..., BPPARAM = bpparam()) {
+		atomindex(object) <- flash(atomindex(object), ..., BPPARAM=BPPARAM)
+		atomdata(object) <- flash(atomdata(object), ..., BPPARAM=BPPARAM)
+		object
+	})
+
 setMethod("atomdata", "sparse_arr",
 	function(object, i = NULL, ...)
 	{
@@ -345,7 +424,7 @@ setMethod("domain", "sparse_arr", function(x) x@domain)
 setReplaceMethod("domain", "sparse_arr", function(x, value) {
 	x@domain <- value
 	if ( length(x@dim) > 2L )
-		stop("more than 2 dimensions are not supported yet")
+		matter_error("more than 2 dimensions are not supported yet")
 	if ( length(x@dim) == 1L ) {
 		x@dim <- length(x@domain)
 		x@dimnames <- NULL
@@ -379,7 +458,7 @@ setReplaceMethod("sampler", "sparse_arr", function(object, ..., value) {
 setMethod("nnzero", "sparse_arr",
 	function(x, na.counted = NA) {
 		if ( !missing(na.counted) )
-			warning("argument 'na.counted' will be ignored")
+			matter_warn("argument 'na.counted' will be ignored")
 		if ( is_sparse_arr_LIL(x) ) {
 			sum(lengths(x@data))
 		} else {
@@ -441,9 +520,10 @@ convert_sparse_arr_LIL <- function(x) {
 		x
 }
 
-subset_sparse_arr_elts <- function(x, i = NULL) {
+subset_sparse_arr_elts <- function(x, i = NULL)
+{
 	if ( !is(x, "sparse_vec") )
-		stop("linear endomorphic subsetting only supported for sparse vectors")
+		matter_error("linear endomorphic subsetting only supported for sparse vectors")
 	if ( is.null(i) )
 		return(x)
 	if ( is.null(domain(x)) ) {
@@ -452,21 +532,25 @@ subset_sparse_arr_elts <- function(x, i = NULL) {
 		domain <- domain(x)[i]
 	}
 	domain(x) <- domain
+	x@ops <- subset_ops(x@ops, list(i))
 	if ( validObject(x) )
 		x
 }
 
-get_sparse_arr_elts <- function(x, i = NULL) {
+get_sparse_arr_elts <- function(x, i = NULL)
+{
 	y <- .Call(C_getSparseArray, x, i, PACKAGE="matter")
 	y <- set_names(y, names(x), i)
 	y
 }
 
-set_sparse_arr_elts <- function(x, i = NULL, value = 0) {
-	stop("sparse array assignment is not supported yet") # TODO
+set_sparse_arr_elts <- function(x, i = NULL, value = 0)
+{
+	matter_error("sparse array assignment is not supported") # TODO?
 }
 
-subset_sparse_mat_submatrix <- function(x, i = NULL, j = NULL) {
+subset_sparse_mat_submatrix <- function(x, i = NULL, j = NULL)
+{
 	if ( x@transpose ) {
 		ii <- j
 		jj <- i
@@ -500,11 +584,13 @@ subset_sparse_mat_submatrix <- function(x, i = NULL, j = NULL) {
 	ncol <- if (is.null(j)) ncol(x) else length(j)
 	x@dim <- c(nrow, ncol)
 	x@dimnames <- subset_dimnames(x@dimnames, list(i, j))
+	x@ops <- subset_ops(x@ops, list(i, j))
 	if ( validObject(x) )
 		x
 }
 
-get_sparse_mat_submatrix <- function(x, i = NULL, j = NULL, drop = FALSE) {
+get_sparse_mat_submatrix <- function(x, i = NULL, j = NULL, drop = FALSE)
+{
 	y <- .Call(C_getSparseMatrix, x, i, j)
 	y <- set_dimnames(y, dimnames(x), list(i, j))
 	if ( drop )
@@ -512,18 +598,19 @@ get_sparse_mat_submatrix <- function(x, i = NULL, j = NULL, drop = FALSE) {
 	y
 }
 
-set_sparse_mat_submatrix <- function(x, i = NULL, j = NULL, value = 0) {
-	stop("sparse array assignment is not implemented yet") # TODO
+set_sparse_mat_submatrix <- function(x, i = NULL, j = NULL, value = 0)
+{
+	matter_error("sparse array assignment is not supported") # TODO?
 }
 
 setMethod("cbind2", c("sparse_mat", "sparse_mat"),
 	function(x, y, ...) {
 		if ( nrow(x) != nrow(y) )
-			stop("number of rows of matrices must match")
+			matter_error("number of rows of matrices must match")
 		if ( rowMaj(x) || rowMaj(y) )
-			stop("can't cbind row-major matrices")
+			matter_error("can't cbind row-major matrices")
 		if ( !is.null(x@ops) )
-			warning("deferred operations will be dropped")
+			matter_warn("deferred operations will be dropped")
 		if ( is.null(x@pointers) && is.null(y@pointers) ) {
 			pointers <- NULL
 		} else {
@@ -545,11 +632,11 @@ setMethod("cbind2", c("sparse_mat", "sparse_mat"),
 setMethod("rbind2", c("sparse_mat", "sparse_mat"),
 	function(x, y, ...) {
 		if ( ncol(x) != ncol(y) )
-			stop("number of columns of matrices must match")
+			matter_error("number of columns of matrices must match")
 		if ( !x@transpose || !y@transpose )
-			stop("can't rbind column-major matrices")
+			matter_error("can't rbind column-major matrices")
 		if ( !is.null(x@ops) )
-			warning("deferred operations will be dropped")
+			matter_warn("deferred operations will be dropped")
 		if ( is.null(x@pointers) && is.null(y@pointers) ) {
 			pointers <- NULL
 		} else {
@@ -577,17 +664,19 @@ setMethod("[", c(x = "sparse_arr"),
 		narg <- nargs() - 1L - !missing(drop)
 		if ( (narg == 1L && !missing(i)) || is.null(dim(x)) ) {
 			i <- as_subscripts(i, x)
-			if ( is_nil(drop) ) {
+			if ( is_null_or_na(drop) ) {
 				subset_sparse_arr_elts(x, i)
 			} else {
 				get_sparse_arr_elts(x, i)
 			}
 		} else {
 			if ( narg != 1L && narg != length(dim(x)) )
-				stop("incorrect number of dimensions")
+				matter_error("incorrect number of dimensions")
+			if ( missing(i) && missing(j) && missing(drop) )
+				drop <- FALSE
 			i <- as_row_subscripts(i, x)
 			j <- as_col_subscripts(j, x)
-			if ( is_nil(drop) ) {
+			if ( is_null_or_na(drop) ) {
 				subset_sparse_mat_submatrix(x, i, j)
 			} else {
 				get_sparse_mat_submatrix(x, i, j, drop)
@@ -604,7 +693,7 @@ setReplaceMethod("[",
 			set_sparse_arr_elts(x, i, value)
 		} else {
 			if ( narg != 1L && narg != length(dim(x)) )
-				stop("incorrect number of dimensions")
+				matter_error("incorrect number of dimensions")
 			i <- as_row_subscripts(i, x)
 			j <- as_col_subscripts(j, x)
 			set_sparse_mat_submatrix(x, i, j, value)
@@ -623,6 +712,12 @@ setMethod("Arith", c(e1 = "vector", e2 = "sparse_arr"),
 setMethod("Arith", c(e1 = "array", e2 = "sparse_arr"),
 	function(e1, e2) register_op(e2, .Generic, e1, TRUE))
 
+setMethod("exp", "sparse_arr", function(x) register_op(x, "exp"))
+setMethod("log", "sparse_arr", function(x) register_op(x, "log"))
+setMethod("log2", "sparse_arr", function(x) register_op(x, "log2"))
+setMethod("log10", "sparse_arr", function(x) register_op(x, "log10"))
+setMethod("log1p", "sparse_arr", function(x) register_op(x, "log1p"))
+
 setMethod("rowMaj", "sparse_arr", function(x)
 {
 	isTRUE(x@transpose)
@@ -630,13 +725,21 @@ setMethod("rowMaj", "sparse_arr", function(x)
 
 setMethod("t", "sparse_arr", function(x)
 {
+	if ( length(dim(x)) > 2L )
+		matter_error("argument is not a matrix")
 	x@transpose <- !x@transpose
 	x@dim <- rev(x@dim)
 	x@dimnames <- rev(x@dimnames)
+	x@ops <- t_ops(x@ops)
 	if ( validObject(x) )
 		x
 })
 
+setMethod("t", "sparse_vec", function(x)
+{
+	t(as(x, "sparse_mat"))
+})
+
 setMethod("%*%", c("sparse_mat", "vector"), function(x, y)
 {
 	if ( rowMaj(x) ) {
