@@ -3,6 +3,49 @@
 
 extern "C" {
 
+// Shared memory
+//--------------
+
+SEXP createSharedMemory(SEXP name)
+{
+	return Rf_ScalarLogical(create_shared_memory_obj(CHAR(Rf_asChar(name))));
+}
+
+SEXP removeSharedMemory(SEXP name)
+{
+	return Rf_ScalarLogical(remove_shared_memory_obj(CHAR(Rf_asChar(name))));
+}
+
+SEXP detectSharedMemory(SEXP names)
+{
+	SEXP result;
+	PROTECT(result = Rf_allocVector(LGLSXP, LENGTH(names)));
+	for ( index_t i = 0; i < LENGTH(names); i++ ) {
+		SEXP name = STRING_ELT(names, i);
+		LOGICAL(result)[i] = detect_shared_memory_obj(CHAR(name));
+	}
+	UNPROTECT(1);
+	return result;
+}
+
+SEXP sizeofSharedMemory(SEXP names)
+{
+	SEXP result;
+	PROTECT(result = Rf_allocVector(REALSXP, LENGTH(names)));
+	for ( index_t i = 0; i < LENGTH(names); i++ ) {
+		SEXP name = STRING_ELT(names, i);
+		REAL(result)[i] = sizeof_shared_memory_obj(CHAR(name));
+	}
+	UNPROTECT(1);
+	return result;
+}
+
+SEXP resizeSharedMemory(SEXP name, SEXP value)
+{
+	return Rf_ScalarReal(resize_shared_memory_obj(CHAR(Rf_asChar(name)),
+		static_cast<size_t>(Rf_asReal(value))));
+}
+
 // Search, sort and select
 //------------------------
 
@@ -236,6 +279,31 @@ SEXP knnSearch(SEXP x, SEXP data, SEXP left_child, SEXP right_child,
 	return result;
 }
 
+SEXP knnSelfSearch(SEXP x, SEXP left_child, SEXP right_child,
+	SEXP knn, SEXP metric, SEXP p)
+{
+	size_t k = Rf_ncols(x);
+	size_t n = Rf_nrows(x);
+	SEXP result;
+	PROTECT(result = Rf_allocMatrix(INTSXP, n, Rf_asInteger(knn)));
+	switch(TYPEOF(x)) {
+		case INTSXP:
+			do_knn_self_search(INTEGER(result), INTEGER(x), k, n,
+				INTEGER(left_child), INTEGER(right_child),
+				Rf_asInteger(knn), Rf_asInteger(metric), Rf_asReal(p), true);
+			break;
+		case REALSXP:
+			do_knn_self_search(INTEGER(result), REAL(x), k, n,
+				INTEGER(left_child), INTEGER(right_child),
+				Rf_asInteger(knn), Rf_asInteger(metric), Rf_asReal(p), true);
+			break;
+		default:
+			Rf_error("unsupported data type");
+	}
+	UNPROTECT(1);
+	return result;
+}
+
 // Distance
 //----------
 
@@ -748,7 +816,7 @@ SEXP guidedFilter(SEXP x, SEXP g, SEXP width,
 	return result;
 }
 
-SEXP iCorr(SEXP x, SEXP y)
+SEXP iCor(SEXP x, SEXP y)
 {
 	switch(TYPEOF(x)) {
 		case INTSXP:
@@ -1174,18 +1242,24 @@ SEXP Approx1(SEXP xi, SEXP x, SEXP y,
 SEXP meanFilter2(SEXP x, SEXP width)
 {
 	SEXP result;
-	PROTECT(result = Rf_allocMatrix(REALSXP, Rf_nrows(x), Rf_ncols(x)));
-	switch(TYPEOF(x)) {
-		case INTSXP:
-			mean_filter2(INTEGER(x), Rf_nrows(x), Rf_ncols(x),
-				Rf_asInteger(width), REAL(result));
-			break;
-		case REALSXP:
-			mean_filter2(REAL(x), Rf_nrows(x), Rf_ncols(x),
-				Rf_asInteger(width), REAL(result));
-			break;
-		default:
-			Rf_error("unsupported data type");
+	PROTECT(result = Rf_allocArray(REALSXP, Rf_getAttrib(x, R_DimSymbol)));
+	size_t n = Rf_nrows(x) * Rf_ncols(x);
+	int nchannels = XLENGTH(x) / n;
+	for ( int i = 0; i < nchannels; i++ )
+	{
+		size_t j = n * i;
+		switch(TYPEOF(x)) {
+			case INTSXP:
+				mean_filter2(INTEGER(x) + j, Rf_nrows(x), Rf_ncols(x),
+					Rf_asInteger(width), REAL(result) + j);
+				break;
+			case REALSXP:
+				mean_filter2(REAL(x) + j, Rf_nrows(x), Rf_ncols(x),
+					Rf_asInteger(width), REAL(result) + j);
+				break;
+			default:
+				Rf_error("unsupported data type");
+		}
 	}
 	UNPROTECT(1);
 	return result;
@@ -1196,18 +1270,24 @@ SEXP linearFilter2(SEXP x, SEXP weights)
 	SEXP result;
 	if ( Rf_nrows(weights) != Rf_ncols(weights) )
 		Rf_error("weights must be a square matrix");
-	PROTECT(result = Rf_allocMatrix(REALSXP, Rf_nrows(x), Rf_ncols(x)));
-	switch(TYPEOF(x)) {
-		case INTSXP:
-			linear_filter2(INTEGER(x), Rf_nrows(x), Rf_ncols(x),
-				REAL(weights), Rf_nrows(weights), REAL(result));
-			break;
-		case REALSXP:
-			linear_filter2(REAL(x), Rf_nrows(x), Rf_ncols(x),
-				REAL(weights), Rf_nrows(weights), REAL(result));
-			break;
-		default:
-			Rf_error("unsupported data type");
+	PROTECT(result = Rf_allocArray(REALSXP, Rf_getAttrib(x, R_DimSymbol)));
+	size_t n = Rf_nrows(x) * Rf_ncols(x);
+	int nchannels = XLENGTH(x) / n;
+	for ( int i = 0; i < nchannels; i++ )
+	{
+		size_t j = n * i;
+		switch(TYPEOF(x)) {
+			case INTSXP:
+				linear_filter2(INTEGER(x) + j, Rf_nrows(x), Rf_ncols(x),
+					REAL(weights), Rf_nrows(weights), REAL(result) + j);
+				break;
+			case REALSXP:
+				linear_filter2(REAL(x) + j, Rf_nrows(x), Rf_ncols(x),
+					REAL(weights), Rf_nrows(weights), REAL(result) + j);
+				break;
+			default:
+				Rf_error("unsupported data type");
+		}
 	}
 	UNPROTECT(1);
 	return result;
@@ -1217,20 +1297,26 @@ SEXP bilateralFilter2(SEXP x, SEXP width,
 	SEXP sddist, SEXP sdrange, SEXP spar)
 {
 	SEXP result;
-	PROTECT(result = Rf_allocMatrix(REALSXP, Rf_nrows(x), Rf_ncols(x)));
-	switch(TYPEOF(x)) {
-		case INTSXP:
-			bilateral_filter2(INTEGER(x), Rf_nrows(x), Rf_ncols(x),
-				Rf_asInteger(width), Rf_asReal(sddist), Rf_asReal(sdrange),
-				Rf_asReal(spar), REAL(result));
-			break;
-		case REALSXP:
-			bilateral_filter2(REAL(x), Rf_nrows(x), Rf_ncols(x),
-				Rf_asInteger(width), Rf_asReal(sddist), Rf_asReal(sdrange),
-				Rf_asReal(spar), REAL(result));
-			break;
-		default:
-			Rf_error("unsupported data type");
+	PROTECT(result = Rf_allocArray(REALSXP, Rf_getAttrib(x, R_DimSymbol)));
+	size_t n = Rf_nrows(x) * Rf_ncols(x);
+	int nchannels = XLENGTH(x) / n;
+	for ( int i = 0; i < nchannels; i++ )
+	{
+		size_t j = n * i;
+		switch(TYPEOF(x)) {
+			case INTSXP:
+				bilateral_filter2(INTEGER(x) + j, Rf_nrows(x), Rf_ncols(x),
+					Rf_asInteger(width), Rf_asReal(sddist), Rf_asReal(sdrange),
+					Rf_asReal(spar), REAL(result) + j);
+				break;
+			case REALSXP:
+				bilateral_filter2(REAL(x) + j, Rf_nrows(x), Rf_ncols(x),
+					Rf_asInteger(width), Rf_asReal(sddist), Rf_asReal(sdrange),
+					Rf_asReal(spar), REAL(result) + j);
+				break;
+			default:
+				Rf_error("unsupported data type");
+		}
 	}
 	UNPROTECT(1);
 	return result;
@@ -1240,20 +1326,26 @@ SEXP diffusionFilter2(SEXP x, SEXP niter,
 	SEXP kappa, SEXP rate, SEXP method)
 {
 	SEXP result;
-	PROTECT(result = Rf_allocMatrix(REALSXP, Rf_nrows(x), Rf_ncols(x)));
-	switch(TYPEOF(x)) {
-		case INTSXP:
-			diffusion_filter2(INTEGER(x), Rf_nrows(x), Rf_ncols(x),
-				Rf_asInteger(niter), Rf_asReal(kappa), Rf_asReal(rate),
-				Rf_asInteger(method), REAL(result));
-			break;
-		case REALSXP:
-			diffusion_filter2(REAL(x), Rf_nrows(x), Rf_ncols(x),
-				Rf_asInteger(niter), Rf_asReal(kappa), Rf_asReal(rate),
-				Rf_asInteger(method), REAL(result));
-			break;
-		default:
-			Rf_error("unsupported data type");
+	PROTECT(result = Rf_allocArray(REALSXP, Rf_getAttrib(x, R_DimSymbol)));
+	size_t n = Rf_nrows(x) * Rf_ncols(x);
+	int nchannels = XLENGTH(x) / n;
+	for ( int i = 0; i < nchannels; i++ )
+	{
+		size_t j = n * i;
+		switch(TYPEOF(x)) {
+			case INTSXP:
+				diffusion_filter2(INTEGER(x) + j, Rf_nrows(x), Rf_ncols(x),
+					Rf_asInteger(niter), Rf_asReal(kappa), Rf_asReal(rate),
+					Rf_asInteger(method), REAL(result) + j);
+				break;
+			case REALSXP:
+				diffusion_filter2(REAL(x) + j, Rf_nrows(x), Rf_ncols(x),
+					Rf_asInteger(niter), Rf_asReal(kappa), Rf_asReal(rate),
+					Rf_asInteger(method), REAL(result) + j);
+				break;
+			default:
+				Rf_error("unsupported data type");
+		}
 	}
 	UNPROTECT(1);
 	return result;
@@ -1265,18 +1357,24 @@ SEXP guidedFilter2(SEXP x, SEXP g, SEXP width,
 	SEXP result;
 	if ( Rf_nrows(x) != Rf_nrows(g) || Rf_ncols(x) != Rf_ncols(g) )
 		Rf_error("signal and guide must have the same dimensions");
-	PROTECT(result = Rf_allocMatrix(REALSXP, Rf_nrows(x), Rf_ncols(x)));
-	switch(TYPEOF(x)) {
-		case INTSXP:
-			guided_filter2(INTEGER(x), INTEGER(g), Rf_nrows(x), Rf_ncols(x),
-				Rf_asInteger(width), Rf_asReal(sdreg), REAL(result));
-			break;
-		case REALSXP:
-			guided_filter2(REAL(x), REAL(g), Rf_nrows(x), Rf_ncols(x),
-				Rf_asInteger(width), Rf_asReal(sdreg), REAL(result));
-			break;
-		default:
-			Rf_error("unsupported data type");
+	PROTECT(result = Rf_allocArray(REALSXP, Rf_getAttrib(x, R_DimSymbol)));
+	size_t n = Rf_nrows(x) * Rf_ncols(x);
+	int nchannels = XLENGTH(x) / n;
+	for ( int i = 0; i < nchannels; i++ )
+	{
+		size_t j = n * i;
+		switch(TYPEOF(x)) {
+			case INTSXP:
+				guided_filter2(INTEGER(x) + j, INTEGER(g) + j, Rf_nrows(x), Rf_ncols(x),
+					Rf_asInteger(width), Rf_asReal(sdreg), REAL(result) + j);
+				break;
+			case REALSXP:
+				guided_filter2(REAL(x) + j, REAL(g) + j, Rf_nrows(x), Rf_ncols(x),
+					Rf_asInteger(width), Rf_asReal(sdreg), REAL(result) + j);
+				break;
+			default:
+				Rf_error("unsupported data type");
+		}
 	}
 	UNPROTECT(1);
 	return result;
@@ -1285,19 +1383,22 @@ SEXP guidedFilter2(SEXP x, SEXP g, SEXP width,
 SEXP histEq(SEXP x, SEXP nbins)
 {
 	SEXP result;
-	if ( Rf_isMatrix(x) )
-		PROTECT(result = Rf_allocMatrix(REALSXP, Rf_nrows(x), Rf_ncols(x)));
-	else
-		PROTECT(result = Rf_allocVector(REALSXP, XLENGTH(x)));
-	switch(TYPEOF(x)) {
-		case INTSXP:
-			histeq(INTEGER(x), XLENGTH(x), Rf_asInteger(nbins), REAL(result));
-			break;
-		case REALSXP:
-			histeq(REAL(x), XLENGTH(x), Rf_asInteger(nbins), REAL(result));
-			break;
-		default:
-			Rf_error("unsupported data type");
+	PROTECT(result = Rf_allocArray(REALSXP, Rf_getAttrib(x, R_DimSymbol)));
+	size_t n = Rf_nrows(x) * Rf_ncols(x);
+	int nchannels = XLENGTH(x) / n;
+	for ( int i = 0; i < nchannels; i++ )
+	{
+		size_t j = n * i;
+		switch(TYPEOF(x)) {
+			case INTSXP:
+				histeq(INTEGER(x) + j, n, Rf_asInteger(nbins), REAL(result) + j);
+				break;
+			case REALSXP:
+				histeq(REAL(x) + j, n, Rf_asInteger(nbins), REAL(result) + j);
+				break;
+			default:
+				Rf_error("unsupported data type");
+		}
 	}
 	UNPROTECT(1);
 	return result;
@@ -1306,21 +1407,49 @@ SEXP histEq(SEXP x, SEXP nbins)
 SEXP adaptHistEq(SEXP x, SEXP width, SEXP clip, SEXP nbins)
 {
 	SEXP result;
-	PROTECT(result = Rf_allocMatrix(REALSXP, Rf_nrows(x), Rf_ncols(x)));
+	PROTECT(result = Rf_allocArray(REALSXP, Rf_getAttrib(x, R_DimSymbol)));
+	size_t n = Rf_nrows(x) * Rf_ncols(x);
+	int nchannels = XLENGTH(x) / n;
+	for ( int i = 0; i < nchannels; i++ )
+	{
+		size_t j = n * i;
+		switch(TYPEOF(x)) {
+			case INTSXP:
+				adapt_histeq(INTEGER(x) + j, Rf_nrows(x), Rf_ncols(x), Rf_asInteger(width),
+					Rf_asReal(clip), Rf_asInteger(nbins), REAL(result) + j);
+				break;
+			case REALSXP:
+				adapt_histeq(REAL(x) + j, Rf_nrows(x), Rf_ncols(x), Rf_asInteger(width),
+					Rf_asReal(clip), Rf_asInteger(nbins), REAL(result) + j);
+				break;
+			default:
+				Rf_error("unsupported data type");
+		}
+	}
+	UNPROTECT(1);
+	return result;
+}
+
+SEXP localMaximaKNN(SEXP x, SEXP neighbors)
+{
+	SEXP ans;
+	if ( LENGTH(x) != Rf_nrows(neighbors) )
+		Rf_error("x and neighbors must have the same extent");
+	PROTECT(ans = Rf_allocVector(LGLSXP, LENGTH(x)));
 	switch(TYPEOF(x)) {
 		case INTSXP:
-			adapt_histeq(INTEGER(x), Rf_nrows(x), Rf_ncols(x), Rf_asInteger(width),
-				Rf_asReal(clip), Rf_asInteger(nbins), REAL(result));
+			local_maxima_knn(INTEGER(x), LENGTH(x), Rf_ncols(neighbors),
+				INTEGER(neighbors), LOGICAL(ans));
 			break;
 		case REALSXP:
-			adapt_histeq(REAL(x), Rf_nrows(x), Rf_ncols(x), Rf_asInteger(width),
-				Rf_asReal(clip), Rf_asInteger(nbins), REAL(result));
+			local_maxima_knn(REAL(x), LENGTH(x), Rf_ncols(neighbors),
+				INTEGER(neighbors), LOGICAL(ans));
 			break;
 		default:
 			Rf_error("unsupported data type");
 	}
 	UNPROTECT(1);
-	return result;
+	return ans;
 }
 
 SEXP Approx2(SEXP xi, SEXP yi, SEXP xy, SEXP z,
